from mpi4py import MPI
import numpy as np
import matplotlib.pyplot as plt
import time
from funcs import *


comm = MPI.COMM_WORLD
size= comm.Get_size(); me = comm.Get_rank()
pid_main = 0; idle = 0; pia = 0.; n=0; ndom=0

prey_eat_radius = 1 
pred_eat_radius = 10


def iter_step_worker():
### WORKER PROCESS
   
    ## receive data from main process
    prey = comm.recv(source=0, tag=11)
    prey_energy = comm.recv(source=0, tag=12)
    prey_sense = comm.recv(source=0, tag=13)
    prey_speed = comm.recv(source=0, tag=14)
    food = comm.recv(source=0, tag=15)
    pred = comm.recv(source=0, tag=16)
    pred_energy = comm.recv(source=0, tag=17)
    pred_sense = comm.recv(source=0, tag=18)
    pred_speed = comm.recv(source=0, tag=19)
    boxes = comm.recv(source=0, tag=31)
    sub_boxes = comm.recv(source=0, tag=32)      
    l = len(boxes)
    l = l//(size-1)
    my_idx = (me-1)*l
    if me == size-1:
        my_boxes = boxes[(me-1)*l:]
    else:
        my_boxes = boxes[(me-1)*l:l*me]

            
    my_prey = []
    my_prey_energy = [] 
    my_prey_speed = []
    my_prey_sense = []
    my_pred = []
    my_pred_energy = []
    my_pred_speed = [] 
    my_pred_sense = []
    my_prey_idx = []
    my_pred_idx = []
           
    prey_prey_sensed = []
    prey_pred_sensed = []
    prey_food_sensed = []
    pred_prey_sensed = []
    pred_pred_sensed = []
    pred_food_sensed = []
    
    ## for each creature find everything in it's sensing distance     
    for box_idx in range(len(my_boxes)):
        box_sub_box_idxs = get_sub_boxes(my_idx+box_idx, int(np.sqrt(len(boxes))), int(np.sqrt(len(sub_boxes))/np.sqrt(len(boxes))))
        for my_sub_box_idx in box_sub_box_idxs:
            bordering_boxes = get_sb_layers(my_sub_box_idx, int(np.sqrt(len(boxes))),int(np.sqrt(len(sub_boxes))/np.sqrt(len(boxes))))
            for i, species in enumerate(sub_boxes[my_sub_box_idx][:2]):

                for j,blob in enumerate(species):
                    if i == 0:
                        my_prey.append(blob)
                        my_prey_energy.append(prey_energy[blob[2]])
                        my_prey_speed.append(prey_speed[blob[2]])
                        my_prey_sense.append(prey_sense[blob[2]])
                    elif i == 1:
                        my_pred.append(blob)
                        my_pred_energy.append(pred_energy[blob[2]])
                        my_pred_speed.append(pred_speed[blob[2]])
                        my_pred_sense.append(pred_sense[blob[2]])
                    near_prey = []
                    near_pred = []
                    near_food = []
                    pred_sensed = False
                    prey_sensed = False
                    food_sensed = False
                    for layer in bordering_boxes:
                        if prey_sensed and pred_sensed and food_sensed:
                            break
                        for layer_box_idx in layer:
                            for prey_it in sub_boxes[layer_box_idx][0]:
                                if prey_sensed:
                                    break
                                if len(layer) == 1:
                                    if i == 0:
                                        if prey_it[2] == blob[2]:
                                            continue
                                near_prey.append(prey_it)
                                
                                
                            for pred_it in sub_boxes[layer_box_idx][1]:
                                if pred_sensed:
                                    break
                                if len(layer) == 1:
                                    if i == 1:
                                        if pred_it[1] == blob[1]:
                                            continue
                                near_pred.append(pred_it)
                            
                            for food_it in sub_boxes[layer_box_idx][2]:
                                if food_sensed:
                                    break
                                near_food.append(food_it)
                                                              
                        if len(near_prey)>0: prey_sensed = True
                        if len(near_pred)>0: pred_sensed = True
                        if len(near_food)>0: food_sensed = True
                        
                    nearest_prey = get_nearest(blob, near_prey)
                    nearest_pred = get_nearest(blob, near_pred)
                    nearest_food = get_nearest(blob, near_food)
                    
                    if i == 0 and nearest_prey[0] > prey_sense[blob[2]]:
                        nearest_prey = [np.inf, None, None]
                    elif i == 1 and nearest_prey[0] > pred_sense[blob[2]]:
                        nearest_prey = [np.inf, None, None]
                    
                    if i == 0 and nearest_pred[0] > prey_sense[blob[2]]:
                        nearest_pred = [np.inf, None, None]
                    elif i == 1 and nearest_pred[0] > pred_sense[blob[2]]:
                        nearest_pred = [np.inf, None, None]
                    
                    if i == 0 and nearest_food[0] > prey_sense[blob[2]]:
                        nearest_food = [np.inf, None, None]                   
                    elif i == 1 and nearest_food[0] > pred_sense[blob[2]]:
                        nearest_food = [np.inf, None, None]
                        
                    if i == 0:
                        prey_prey_sensed.append(nearest_prey)
                        prey_pred_sensed.append(nearest_pred)
                        prey_food_sensed.append(nearest_food)
                    elif i == 1:
                        pred_prey_sensed.append(nearest_prey)
                        pred_pred_sensed.append(nearest_pred)
                        pred_food_sensed.append(nearest_food)

                
                ## move
    if len(my_prey)>0:            
        updt_prey, updt_prey_energy = prey_move(my_prey,prey, pred, my_prey_energy, my_prey_sense, my_prey_speed, food, prey_prey_sensed, prey_pred_sensed,prey_food_sensed)
    else:
        updt_prey,updt_prey_energy = my_prey, my_prey_energy
    if len(my_pred)> 0:
        updt_pred, updt_pred_energy = pred_move(my_pred,prey, pred, my_pred_energy, my_pred_sense, my_pred_speed, food, pred_prey_sensed, pred_pred_sensed,pred_food_sensed)
    else:
        updt_pred,updt_pred_energy = my_pred, my_pred_energy
    
    for i,_ in enumerate(updt_prey):
        my_prey_idx.append(updt_prey[i][2])
        updt_prey[i] = updt_prey[i][:2]
    for i,_ in enumerate(updt_pred):
        my_pred_idx.append(updt_pred[i][2])
        updt_pred[i] = updt_pred[i][:2]
    

    ## send all generated data back	
    comm.send(updt_prey, dest=0, tag=21)
    comm.send(updt_prey_energy, dest=0, tag=22)
    comm.send(updt_pred, dest=0, tag=26)
    comm.send(updt_pred_energy, dest=0, tag=27)
    comm.send(my_prey_idx, dest=0, tag=30)
    comm.send(my_pred_idx, dest=0, tag=31)
    comm.send(prey_food_sensed, dest=0, tag=50)
    comm.send(pred_prey_sensed, dest=0, tag=51)  

### WORKER PROCESS 

def iter_step_main(gen,prey,prey_energy,prey_sense,prey_speed,food,pred,pred_energy,pred_sense,pred_speed,food_eaten,spawns_food):       
### MAIN PROCESS
    x_count = min([lim_x[1]//max(prey_sense), lim_x[1]//max(pred_sense)])
    y_count = min([lim_y[1]//max(prey_sense), lim_y[1]//max(pred_sense)])
    x_len = (lim_x[1]/x_count)+1
    y_len = (lim_y[1]/y_count)+1
    
    prey_boxes = get_boxes(x_count,x_len,y_count,y_len,prey)
    pred_boxes = get_boxes(x_count,x_len,y_count,y_len,pred)
    food_boxes = get_boxes(x_count,x_len,y_count,y_len,food)
    boxes, sub_boxes = init_boxes(5,3,prey,pred,food)
    
    
    
    
    ## send data to worker processes
    for i in range(1,size):
        comm.send(prey,dest=i, tag=11)
        comm.send(prey_energy,dest=i, tag=12)
        comm.send(prey_sense,dest=i, tag=13)
        comm.send(prey_speed,dest=i, tag=14)
        comm.send(food,dest=i, tag=15)
        comm.send(pred,dest=i, tag=16)
        comm.send(pred_energy,dest=i, tag=17)
        comm.send(pred_sense,dest=i, tag=18)
        comm.send(pred_speed,dest=i, tag=19)
        comm.send(boxes, dest=i, tag=31)
        comm.send(sub_boxes, dest=i, tag=32)
    ## recollect data from worker processes
    buf = np.array([]).reshape((0,2))
    buf2 = np.array([]).reshape((0))
    buf6 = np.array([]).reshape((0,2))
    buf7 = np.array([]).reshape((0))
    buf30 = np.array([]).reshape((0))
    buf31 = np.array([]).reshape((0))
    buf50 = np.array([]).reshape((0,3))
    buf51 = np.array([]).reshape((0,3))

    for i in range(1,size):
        prey = comm.recv(source=i, tag=21)
        prey_energy = comm.recv(source=i, tag=22)
        pred = comm.recv(source=i, tag=26)
        pred_energy = comm.recv(source=i, tag=27)
        prey_idx = comm.recv(source = i, tag=30)
        pred_idx = comm.recv(source = i, tag=31)
        prey_want_to_eat = comm.recv(source=i, tag=50)
        pred_want_to_eat = comm.recv(source=i, tag=51)

        if len(prey) > 0:
            buf = np.concatenate((buf,prey),axis=0)
            buf2 = np.concatenate((buf2,prey_energy),axis=0)
            buf30 = np.concatenate((buf30,prey_idx),axis=0)
            buf50 = np.concatenate((buf50,prey_want_to_eat),axis=0)
        if len(pred) > 0:
            buf6 = np.concatenate((buf6,pred),axis=0)
            buf7 = np.concatenate((buf7,pred_energy),axis=0)
            buf31 = np.concatenate((buf31,pred_idx),axis=0)
            buf51 = np.concatenate((buf51,pred_want_to_eat),axis=0)
    prey = buf
    prey_energy = buf2
    pred = buf6
    pred_energy = buf7
    prey_idx = buf30
    pred_idx = buf31
    prey_want_to_eat = buf50
    pred_want_to_eat = buf51
    
    prey_sort = prey.copy()
    prey_energy_sort = prey_energy.copy()
    prey_wte_sort = prey_want_to_eat.copy()
    for i,v in enumerate(prey_idx):
        prey_sort[int(v)] = prey[i]
        prey_energy_sort[int(v)] = prey_energy[i]
        prey_wte_sort[int(v)] = prey_want_to_eat[i]
    
    pred_sort = pred.copy()
    pred_energy_sort = pred_energy.copy()
    pred_wte_sort = pred_want_to_eat.copy()
    
    for i,v in enumerate(pred_idx):
        pred_sort[int(v)] = pred[i]
        pred_energy_sort[int(v)] = pred_energy[i]
        pred_wte_sort[int(v)] = pred_want_to_eat[i]
    prey = prey_sort
    prey_energy = prey_energy_sort
    prey_want_to_eat = prey_wte_sort
    pred = pred_sort
    pred_energy = pred_energy_sort
    pred_want_to_eat = pred_wte_sort
    
    ## prey and predator eat
    want_to_eat = [prey_want_to_eat,pred_want_to_eat]
    eat = [prey_eat_radius,pred_eat_radius]
    energy = [prey_energy, pred_energy]
    prey_eaten = 0
    food_eaten_it = 0
    for j in [0,1]:      
        can_eat = want_to_eat[j][:,0]<eat[j]
        can_eat_idx = np.arange(len(want_to_eat[j]))[can_eat]
        already_eaten=[]
        for i in can_eat_idx:
            if want_to_eat[j][i,2] in already_eaten:
                continue
            count = np.count_nonzero(want_to_eat[j][i,1]==want_to_eat[j][can_eat_idx,1])
            if count == 1:
                if j == 0 and food[int(want_to_eat[j][i,2])][0] != np.inf:
                    food[int(want_to_eat[j][i,2])] = np.array([np.inf,np.inf])
                    prey_energy[i] += 100
                    food_eaten_it += 1
                elif j == 1 and prey_energy[int(want_to_eat[j][i,2])] > 0:
                    pred_energy[i] += prey_energy[int(want_to_eat[j][i,2])] + 80
                    prey_energy[int(want_to_eat[j][i,2])] = 0
                    prey_eaten += 1

            else:
                decide = np.random.randint(count)
                decide = np.nonzero(want_to_eat[j][i,1]==want_to_eat[j][can_eat_idx,1])[0][decide]
                if j == 0 and food[int(want_to_eat[j][int(decide),1])][0] != np.inf:
                    food[int(want_to_eat[j][int(decide),1])] = np.array([np.inf,np.inf])
                    prey_energy[decide] += 100
                    food_eaten_it += 1
                elif j == 1 and prey_energy[int(want_to_eat[j][i,1])] > 0 :
                    pred_energy[i] += prey_energy[int(want_to_eat[j][i,1])] + 80
                    prey_energy[int(want_to_eat[j][i,1])] = 0
                    prey_eaten += 1
                already_eaten.append(want_to_eat[j][decide,1])
   
    food = np.delete(food,np.nonzero(food[:,0] == np.inf),axis=0)

    
    ## update food and creatures
    prey,prey_energy,prey_sense,prey_speed = check_alive(prey,prey_energy,prey_sense,prey_speed,)
    pred,pred_energy,pred_sense,pred_speed = check_alive(pred,pred_energy,pred_sense,pred_speed,)

    food_eaten = np.concatenate((food_eaten[1:10], [food_eaten_it]))
    food = spawn_food(food,spawns_food,food_eaten[0])

    prey,prey_energy,prey_sense,prey_speed = reproduce(prey,prey_energy,prey_sense,prey_speed,'prey')
    pred,pred_energy,pred_sense,pred_speed = reproduce(pred,pred_energy,pred_sense,pred_speed,'pred')

    return prey,prey_energy,prey_sense,prey_speed,food,pred,pred_energy,pred_sense,pred_speed,food_eaten
